#include "hvif-light.h"

#include <stdint.h>
#include <stdlib.h>

enum
{
	STYLE_TYPE_SOLID_COLOR = 1,
	STYLE_TYPE_GRADIENT = 2,
	STYLE_TYPE_SOLID_COLOR_NO_ALPHA = 3,
	STYLE_TYPE_SOLID_GRAY = 4,
	STYLE_TYPE_SOLID_GRAY_NO_ALPHA = 5,

	SHAPE_TYPE_PATH_SOURCE = 10,

	TRANSFORMER_TYPE_AFFINE = 20,
	TRANSFORMER_TYPE_CONTOUR = 21,
	TRANSFORMER_TYPE_PERSPECTIVE = 22,
	TRANSFORMER_TYPE_STROKE = 23,
};

enum gradients_type
{
	GRADIENT_LINEAR = 0,
	GRADIENT_CIRCULAR,
	GRADIENT_DIAMOND,
	GRADIENT_CONIC,
	GRADIENT_XY,
	GRADIENT_SQRT_XY,
	HVIF_GRADIENT_TYPE_MAX
};

enum
{
	GRADIENT_FLAG_TRANSFORM = 1 << 1,
	GRADIENT_FLAG_NO_ALPHA = 1 << 2,
	GRADIENT_FLAG_16_BIT_COLORS = 1 << 3, /* not yet used */
	GRADIENT_FLAG_GRAYS = 1 << 4,
};

enum
{
	PATH_FLAG_CLOSED = 1 << 1,
	PATH_FLAG_USES_COMMANDS = 1 << 2,
	PATH_FLAG_NO_CURVES = 1 << 3,
};

enum
{
	PATH_COMMAND_H_LINE = 0,
	PATH_COMMAND_V_LINE = 1,
	PATH_COMMAND_LINE = 2,
	PATH_COMMAND_CURVE = 3,
};

enum
{
	SHAPE_FLAG_TRANSFORM = 1 << 1,
	SHAPE_FLAG_HINTING = 1 << 2,
	SHAPE_FLAG_LOD_SCALE = 1 << 3,
	SHAPE_FLAG_HAS_TRANSFORMERS = 1 << 4,
	SHAPE_FLAG_TRANSLATION = 1 << 5,
};

#define ERROR_RESULT(E) \
	(hvif_result) { .status = E }
#define SUCCESS_RESULT(I) \
	(hvif_result) { .status = SUCCESS, .image = I }

/* An affine transformation. Ugly until I understand better how they are used */
typedef struct hvif_matrix hvif_matrix;
struct hvif_matrix
{
	double v1;
	double v2;
	double v3;
	double v4;
	double v5;
	double v6;
};

/* TODO: This is most likely wrong */
#define ID_TRANSFORMATION \
	(hvif_matrix) { 1.0, 0.0, 0.0, 0.0, 1.0, 0.0 }

#define hvif_read(V, BUFFER) \
	V = _Generic((V), uint32_t \
							 : hvif_read_uint32, float \
							 : hvif_read_float, hvif_matrix \
							 : hvif_read_hvif_matrix)(BUFFER)
static inline uint32_t
hvif_read_uint32(char buffer[static 1])
{
	return (buffer[0] << 0) | (buffer[1] << 8) | (buffer[2] << 16) |
		(buffer[3] << 24);
}
static inline float
hvif_read_float(char buffer[static 1])
{
	int shortValue = (buffer[0] << 16) | (buffer[1] << 8) | buffer[2];
	int sign = (shortValue & 0x800000) >> 23;
	int exponent = ((shortValue & 0x7e0000) >> 17) - 32;
	int mantissa = (shortValue & 0x01ffff) << 6;

	if (shortValue == 0)
		return 0.0;
	else {
		uint32_t value = (sign << 31) | ((exponent + 127) << 23) | mantissa;
		return (float)value;
	}
}
static inline hvif_matrix
hvif_read_hvif_matrix(char buffer[static 1])
{
	return (
		hvif_matrix){ hvif_read_float(buffer),      hvif_read_float(&buffer[3]),
									hvif_read_float(&buffer[6]),  hvif_read_float(&buffer[9]),
									hvif_read_float(&buffer[12]), hvif_read_float(&buffer[15]) };
}

typedef uint32_t hvif_color; /* 8bit each: alpha, blue, green, red */
#define COLOR_GET_RED(C) (C & 0xff)
#define COLOR_GET_GREEN(C) ((C >> 8) & 0xff)
#define COLOR_GET_BLUE(C) ((C >> 16) & 0xff)
#define COLOR_GET_ALPHA(C) ((C >> 24) & 0xff)
#define _BM8(N) (0xff << N)
#define COLOR_SET_RED(C, V) (C = ((C & (~_BM8(0))) | (V & 0xff)))
#define COLOR_SET_GREEN(C, V) (C = ((C & (~_BM8(8))) | ((V & 0xff) << 8)))
#define COLOR_SET_BLUE(C, V) (C = ((C & (~_BM8(16))) | ((V & 0xff) << 16)))
#define COLOR_SET_ALPHA(C, V) (C = ((C & (~_BM8(24))) | ((V & 0xff) << 24)))
#define COLOR_SET_RGB(C, R, G, B) \
	(C = (C & (~0xffffff)) | (R & 0xff) | ((G & 0xff) << 8) | ((B & 0xff) << 16))

typedef struct hvif_style hvif_style;
struct hvif_style
{
	signed gradient_type;
	uint8_t stops;
	hvif_color* colors;
	float* offsets;
	hvif_matrix transformation;
};

struct hvif_image
{
	uint8_t num_styles;
	hvif_style* styles;
};

hvif_color
decode_color(char* buffer, bool gray, bool alpha)
{
	/* On disk a color is:
	 	   red, green, blue, alpha
	 	 Hence reading this as a (little-endian) uint32 result in the desired
	 	 format. If `alpha` is `false` then only three bytes are used and alpha
	 	 is `0xff`. If gray is `true` only one one byte is used for the three
	 	 colors.
	 */
	hvif_color c = 0;
	if (alpha) {
		if (gray) {
			COLOR_SET_RGB(c, buffer[0], buffer[0], buffer[0]);
			COLOR_SET_ALPHA(c, buffer[1]);
		} else {
			hvif_read(c, buffer);
		}
	} else {
		COLOR_SET_ALPHA(c, 0xff);
		if (gray) {
			COLOR_SET_RGB(c, buffer[0], buffer[0], buffer[0]);
		} else {
			COLOR_SET_RGB(c, buffer[0], buffer[1], buffer[2]);
		}
	}
	return c;
}

hvif_status
read_color_style(hvif_style* style, char* buffer, bool gray, bool alpha)
{
	hvif_color color = decode_color(buffer, gray, alpha);
	style->colors = calloc(1, sizeof(hvif_color));
	style->offsets = calloc(1, sizeof(float));
	if (!style->colors || !style->colors) return ERROR_NOMEM;

	style->colors[0] = color;
	style->offsets[0] = 1.0; /* TODO: Is this ok for solid colors */
	return SUCCESS;
}

hvif_status
read_gradient_style(hvif_style* style, FILE* file, char* style_buffer)
{
	uint8_t gradient_type = style_buffer[0];
	uint8_t gradient_flags = style_buffer[1];
	uint8_t gradient_stops = style_buffer[2];

	if (gradient_type >= HVIF_GRADIENT_TYPE_MAX) return ERROR_STYLE;
	style->gradient_type = gradient_type;

	bool gray = gradient_flags & GRADIENT_FLAG_GRAYS;
	bool alpha = !(gradient_flags & GRADIENT_FLAG_NO_ALPHA);

	/* structure:
		 * if transformable [matrix]
		 * then number of stops entries of the form: [stop][color] */

	hvif_matrix transformation;
	if (gradient_flags & GRADIENT_FLAG_TRANSFORM) {
		/* The transformation matrix is 6 hvif-floats: 18 bytes */
		char buffer[18];
		if (fread(buffer, 1, 2, file) != 2) { return ERROR_EOF; }
		hvif_read(transformation, buffer);
	} else {
		transformation = ID_TRANSFORMATION;
	}
	style->transformation = transformation;

	style->offsets = calloc(gradient_stops, sizeof(float));
	style->colors = calloc(gradient_stops, sizeof(hvif_color));

	char buffer[4];
	for (unsigned i = 0; i < gradient_stops; ++i) {
		if (fread(buffer, 1, 1, file) != 1) { return ERROR_EOF; }
		float offset = buffer[0] / 255.0;
		hvif_color c;
		if (alpha) {
			if (gray) {
				if (fread(buffer, 1, 2, file) != 2) { return ERROR_EOF; }
				c = decode_color(buffer, gray, alpha);
			} else {
				if (fread(buffer, 1, 4, file) != 1) { return ERROR_EOF; }
				c = decode_color(buffer, gray, alpha);
			}
		} else {
			if (gray) {
				if (fread(buffer, 1, 1, file) != 1) { return ERROR_EOF; }
				c = decode_color(buffer, gray, alpha);
			} else {
				if (fread(buffer, 1, 3, file) != 1) { return ERROR_EOF; }
				c = decode_color(buffer, gray, alpha);
			}
		}
		style->offsets[i] = offset;
		style->colors[i] = c;
	}

	return SUCCESS;
}

hvif_status
read_style(FILE* file, hvif_style* style)
{
	uint8_t type;
	if (fread(&type, 1, 1, file) != 1) { return ERROR_STYLE; }
	char buffer[4];
	switch (type) {
		case STYLE_TYPE_SOLID_COLOR:
			if (fread(buffer, 1, 4, file) != 4) { return ERROR_EOF; }
			read_color_style(style, buffer, false, true);
			break;
		case STYLE_TYPE_GRADIENT:
			if (fread(buffer, 1, 3, file) != 3) { return ERROR_EOF; }
			read_gradient_style(style, file, buffer);
			break;
		case STYLE_TYPE_SOLID_COLOR_NO_ALPHA:
			if (fread(buffer, 1, 3, file) != 3) { return ERROR_EOF; }
			read_color_style(style, buffer, false, false);
			break;
		case STYLE_TYPE_SOLID_GRAY:
			if (fread(buffer, 1, 2, file) != 2) { return ERROR_EOF; }
			read_color_style(style, buffer, true, true);
			break;
		case STYLE_TYPE_SOLID_GRAY_NO_ALPHA:
			if (fread(buffer, 1, 1, file) != 1) { return ERROR_EOF; }
			read_color_style(style, buffer, true, false);
			break;
		default: return ERROR_STYLE;
	}
	return SUCCESS;
}

hvif_result
hvif_from_file(FILE* file)
{
	uint32_t const magic = 0x6669636e; /* 'finc' (on disk little endian 'cnif') */

	char magic_buffer[4];
	if (fread(magic_buffer, 1, 4, file) != 4) { return ERROR_RESULT(ERROR_EOF); }
	uint32_t read_magic;
	hvif_read(read_magic, magic_buffer);
	if (read_magic != magic) { return ERROR_RESULT(ERROR_MAGIC); }

	hvif_image* image = calloc(1, sizeof(hvif_image));
	if (!image) return ERROR_RESULT(ERROR_NOMEM);

	uint8_t num_styles;
	if (fread(&num_styles, 1, 1, file) != 1) {
		hvif_free(image);
		return ERROR_RESULT(ERROR_EOF);
	}
	image->styles = calloc(num_styles, sizeof(hvif_style));
	if (!image->styles) {
		hvif_free(image);
		return ERROR_RESULT(ERROR_NOMEM);
	}
	for (unsigned i = 0; i < image->num_styles; ++i) {
		hvif_status s = read_style(file, &image->styles[i]);
		if (s != SUCCESS) {
			hvif_free(image);
			return ERROR_RESULT(s);
		}
	}

	return SUCCESS_RESULT(image);
}

void
hvif_free(hvif_image* image)
{
	for (unsigned i = 0; i < image->num_styles; ++i) {
		free(image->styles[i].colors);
		free(image->styles[i].offsets);
	}
	free(image->styles);
	free(image);
}
